module emop.syntax {
   syntax(X) :-
      logic(X),
      comparisons(X),
      arithmetic(X),
      module X {
         syntax: '\\+' < '=' < infix '+'.
         syntax: prefix '+' < infix ':'.
      },
      level(X),
      functional_logic(X),
      pipe_match(X),
      implicit_do(X).

   logic = module {
      constructor: '.'/2, ','/2, '[]'/0, top/0.

      syntax: infix right associative '.' ';' '->' ',' '>>'.
      syntax: infix ':-'.
      syntax: prefix '\\+' ':'.
      syntax: infix left associative ':'.
      syntax: postfix '.()'.

      syntax: '.' < ':-' < ';' < '->' < ',' < '\\+' < '>>' < prefix ':' = infix ':' = '.()'.
   }.

   comparisons = module {
      syntax: infix left associative '=' '<' '>' '=<' '>=' '\\=' '=='.

      syntax: '=' = '<' = '>' = '=<' = '>=' = '\\=' = '=='.
   }.

   arithmetic = module {
      syntax: infix left associative '+' '-' '*' '/' 'mod' 'quo'.
      syntax: infix right associative '**'.
      syntax: prefix '+' '-'.

      syntax: infix '+' = infix '-' < '*' = '/' = 'mod' = 'quo'.
      syntax: '*' < '**' < prefix '+' = prefix '-'.
   }.

   level = module {
      level: (goal) (goal, goal).
      level: (goal) (goal; goal).
      level: (goal) (goal -> goal).
      level: (goal) (goal >> goal).
      level: (goal) (term:goal).
      level: (goal) 'catch'(goal, term, goal).
      level: (goal) 'call_det'(goal, term).
      level: (goal) (\+ goal).

      level: (term) (goal -> term).
      level: (term) (goal >> term).

      level: (goal, term) p(term*).
   }.

   functional_logic = module {
      notation: (term) (P -> Q) -> (P -> true) >> Q.
      notation: (term) (P -> Q; R) -> (P -> X = Q; X = R) >> X.
      notation: (term) (P; Q) -> (X = P; X = Q) >> X.
      notation: (term) (A = B) -> (X = A, X = B) >> X.
   }.

   pipe_match = module {
      syntax: infix left associative '|' 'match'.
      syntax: infix right associative '=>'.
      syntax: prefix '|' 'match'.

      syntax: '=' < infix '|' = infix 'match' < prefix '|' = prefix 'match' < '>>'.
      syntax: '=>' = '->'.

      notation: (goal, term) (A | match B) -> A match B.
      notation: (goal, term) (A | | B) -> A | B.
      notation: (goal, term) (A match match B) -> A match B.
      notation: (goal, term) (A match | B) -> A | B.

      syntax: infix left associative '&'.

      syntax: '=' < infix '|' < prefix '|' < '&' < '>>'.

      notation: (goal) (A | X) | X: var; X: number -> A = X.
      notation: (term) (A | V) | V: var -> (A = X, X = V) >> X.
      notation: (term) (A | N) | N: number -> (A = N) >> N.
      notation: (goal, term) (A | (P >> Q)) -> (A = X) >> P >> (X | Q).
      notation: (goal, term) (A | (P -> Q)) -> (A = X) >> (P -> X | Q).
      notation: (goal, term) (A | (P -> Q; R)) -> (A = X) >> (P -> X | Q; X | R).
      notation: (goal, term) (A | (P; Q)) -> (A = X) >> (X | P; X | Q).
      notation: (goal, term) (A | X:p(Gamma*)) -> X:p(A, Gamma).
      notation: (goal, term) (A | p(Gamma*)) -> p(A, Gamma).

      notation: (goal, term) (A | (P & Q)) -> (A = X) >> (X | P) >> (X | Q).
      notation: (goal, term) ((P & Q) | A) -> (A = X) >> (P | X) >> (Q | X).


      notation: (goal) A match X | X: var; X: number -> A = X.

      notation: (term) A match V | V: var -> (A = X, X = V) >> X.
      notation: (term) A match N | N: number -> (A = N) >> N.
      notation: (goal, term) A match (P >> Q) -> (A = X, P) >> (X match Q).
      notation: (goal, term) A match (P => Q) -> (A = X) >> (X match P) >> Q.
      notation: (goal, term) A match (P -> Q) -> (A = X) >> (X match P -> Q).
      notation: (goal, term) A match (P -> Q; R) -> (A = X) >> (X match P -> Q; X match R).
      notation: (goal, term) A match (P; Q) -> (A = X) >> (X match P; X match Q).
      notation: (goal, term) A match (P & Q) -> (A = X) >> (X match P) >> (X match Q).
      notation: (goal, term) (P & Q) match A -> (A = X) >> (P match A) >> (Q match A).
      notation: (goal) A match X:p(B*) -> (A = X:p(C*)) >> ((C match B)*).
      notation: (goal) A match p(B*) -> (A = p(C*)) >> ((C match B)*).
      notation: (term) A match X:p(B*) -> (A = X:p(C*)) >> X:p((C match B)*).
      notation: (term) A match p(B*) -> (A = p(C*)) >> p((C match B)*).

      notation: (goal, term) (X | module M [A, Gamma*] { B }) -> (X = A) >> module M [A, Gamma] { B }.
   }.

   implicit_do = module {
      notation: (term) X:f(Gamma*, '*', Delta*) ->
         module { do(A) :- X:f(Gamma*, A, Delta*) }.

      notation: (term) f(Gamma*, '*', Delta*) ->
         module { do(A) :- f(Gamma*, A, Delta*) }.

      notation: (term) X:f(Gamma*, '<', Delta*, '>', Epsilon*) ->
         module { do(A, B) :- X:f(Gamma*, A, Delta*, B, Epsilon*) }.

      notation: (term) f(Gamma*, '<', Delta*, '>', Epsilon*) ->
         module { do(A, B) :- f(Gamma*, A, Delta*, B, Epsilon*) }.

      notation: (term) X:f(Gamma*, '>', Delta*, '<', Epsilon*) ->
         module { do(A, B) :- X:f(Gamma*, B, Delta*, A, Epsilon*) }.

      notation: (term) f(Gamma*, '>', Delta*, '<', Epsilon*) ->
         module { do(A, B) :- f(Gamma*, B, Delta*, A, Epsilon*) }.
   }
}
